﻿
using System;
using System.Collections.Generic;
using System.Threading;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Collections.Generic.Extensions;
using OpcLabs.BaseLib.ComInterop;
using OpcLabs.BaseLib.Runtime;
using OpcLabs.EasyOpc;
using OpcLabs.EasyOpc.DataAccess;
using OpcLabs.EasyOpc.DataAccess.Extensions;
using OpcLabs.EasyOpc.DataAccess.OperationModel;
using OpcLabs.EasyOpcLiveTest.DataAccess.Environment;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.DataAccess
{
    partial class EasyDAClientTestBase
    {

        protected void ChangeItemSubscriptions_Basic(
            [NotNull] ServerDescriptor serverDescriptor,
            [NotNull] DAItemDescriptor itemDescriptor)
        {
            TestSection.Arrange();
            AssertUtilities.WriteLine(serverDescriptor);
            AssertUtilities.WriteLine(itemDescriptor);
            AssertUtilities.WriteLine("");

            TestEnvironment.PrecheckServer(serverDescriptor);

            var senderList = new List<object>();
            var argsList = new List<EasyDAItemChangedEventArgs>();
            int indexSet = -1;
            var completeEvent = new AutoResetEvent(initialState: false);

            void ClientOnItemChanged(object sender, EasyDAItemChangedEventArgs args)
            {
                AssertUtilities.WriteLine(args);
                senderList.Add(sender);
                argsList.Add(args);

                if ((indexSet == -1) &&
                    !(args is null) &&
                    ((args.Exception != null) || ((args.Vtq != null) && !args.Vtq.Quality.IsBad)))
                {
                    completeEvent.Set();
                    indexSet = argsList.Count - 1;
                }
            }

            const int requestedUpdateRate1 = 200;
            const int requestedUpdateRate2 = 100;
            const int testState = 777777777;

            EasyDAClient client = CreateClient();

            TestSection.Act();
            const int waitTimeout1 = 8 * 1000;    // 5 seconds was not always enough
            const int waitTimeout2 = 5 * 1000;

            client.ItemChanged += ClientOnItemChanged;

            int handle = client.SubscribeItem(serverDescriptor, itemDescriptor, requestedUpdateRate1, testState);
            var completed = completeEvent.WaitOne(waitTimeout1);

            client.ChangeItemSubscription(handle, new DAGroupParameters(requestedUpdateRate2));
            AssertUtilities.WaitingPhase(waitTimeout2);

            client.UnsubscribeItem(handle);
            client.ItemChanged -= ClientOnItemChanged;

            TestSection.Assert();

            int count = argsList.Count;
            AssertUtilities.WriteLine($"count: {count}");
            AssertUtilities.WriteLine($"indexSet: {indexSet}");
            Assert.IsTrue(completed, "Not completed.");

            for (int i = 0; i < argsList.Count; i++)
            {
                object sender = senderList[i];
                Assert.IsNotNull(sender);
                Assert.AreSame(client, sender);

                EasyDAItemChangedEventArgs args = argsList[i];
                Assert.IsNotNull(args);

                Assert.AreEqual(testState, args.Arguments.State); // is correct for value types only
                Assert.IsNull(args.Arguments.Callback);

                Assert.IsTrue(
                    (args.Arguments.GroupParameters == new DAGroupParameters(requestedUpdateRate1)) ||
                    (args.Arguments.GroupParameters == new DAGroupParameters(requestedUpdateRate2)));
                Assert.AreEqual(itemDescriptor, args.Arguments.ItemDescriptor);
                Assert.AreEqual(serverDescriptor, args.Arguments.ServerDescriptor);

                if (!client.SupportsTechnology(serverDescriptor.Technology))
                {
                    Assert.IsFalse(args.Succeeded);
                    Assert.IsInstanceOfType(args.Exception, typeof(NotSupportedException));
                    continue;
                }

                Assert.IsNull(args.Exception);
                Assert.AreEqual(null, args.ErrorId);

                DAVtq actualVtq = args.Vtq;
                Assert.IsNotNull(actualVtq);
                AssertUtilities.WriteLine($@"{i}: {actualVtq}");

                object actualValue = actualVtq.Value;

                if ((indexSet == -1) || (i < indexSet))
                {
                    Assert.IsTrue(actualVtq.Quality.IsBad);
                    Assert.IsNull(actualValue);
                }
                else
                {
                    Assert.IsTrue(actualVtq.Quality.IsGood);
                }

                Assert.AreEqual(DateTimeKind.Utc, actualVtq.Timestamp.Kind);
                Assert.AreNotEqual(actualVtq.Timestamp, DateTime.MinValue);
                Assert.AreNotEqual(actualVtq.Timestamp, DateTime.MaxValue);
                Assert.IsTrue(actualVtq.Timestamp >= DateTime.FromFileTimeUtc(0));
                DateTime utcNow = DateTime.UtcNow;
                AssertUtilities.WriteLine($"UtcNow: {utcNow}");
                TimeSpan timeTolerance = TimeSpan.FromSeconds(60);
                Assert.IsTrue(actualVtq.Timestamp <= utcNow + timeTolerance, "actualVtq.Timestamp <= utcNow + timeTolerance");
            }

            Assert.IsTrue(count > indexSet + 5);
        }

        protected void SubscribeItemCatalogTest_Basic(
            [NotNull] ServerDescriptor serverDescriptor,
            [NotNull] DAItemCatalogEntry itemCatalogEntry)
        {
            VarType dataType = itemCatalogEntry.DataType;
            Type expectedType = null;
            if (dataType != VarTypes.Empty)
                expectedType = Cls.GetClsType(VarTypeUtilities.TypeFromVarType(dataType));

            SubscribeItemTest_Basic(serverDescriptor, itemCatalogEntry.Element, useEventPull:false, expectedType);
        }

        protected void SubscribeItemTest_Basic(
            [NotNull] ServerDescriptor serverDescriptor,
            [NotNull] DAItemDescriptor itemDescriptor,
            bool useEventPull,
            [CanBeNull] Type expectedType)
        {
            TestSection.Arrange();
            AssertUtilities.WriteLine(serverDescriptor);
            AssertUtilities.WriteLine(itemDescriptor);
            AssertUtilities.WriteLine(expectedType);
            AssertUtilities.WriteLine("");

            TestEnvironment.PrecheckServer(serverDescriptor);

            var senderList = new List<object>();
            var argsList = new List<EasyDAItemChangedEventArgs>();
            int indexSet = -1;
            var completeEvent = new AutoResetEvent(initialState: false);

            void ClientOnItemChanged(object sender, EasyDAItemChangedEventArgs args)
            {
                AssertUtilities.WriteLine(args);
                senderList.Add(sender);
                argsList.Add(args);

                if ((indexSet == -1) &&
                    !(args is null) && 
                    ((args.Exception != null) || ((args.Vtq != null) && !args.Vtq.Quality.IsBad)))
                {
                    completeEvent.Set();
                    indexSet = argsList.Count - 1;
                }
            }

            const int requestedUpdateRate = 100;
            const int testState = 777777777;

            EasyDAClient client = CreateClient();

            TestSection.Act();
            const int waitTimeout = 8 * 1000;    // 5 seconds was not always enough
            bool completed;
            if (useEventPull)
            {
                client.PullItemChangedQueueCapacity = 1000;
                int handle = client.SubscribeItem(serverDescriptor, itemDescriptor, requestedUpdateRate, testState);
                int startTick = System.Environment.TickCount;
                int endTick = System.Environment.TickCount + waitTimeout;
                do
                {
                    EasyDAItemChangedEventArgs[] argsArray = client.PullMultipleItemChanges(
                        Int32.MaxValue, 
                        Math.Max(endTick - System.Environment.TickCount, 0));
                    argsArray?.ForEach(args => ClientOnItemChanged(client, args));
                } while (System.Environment.TickCount < startTick + waitTimeout);
                completed = completeEvent.WaitOne(0);
                client.UnsubscribeItem(handle);
            }
            else
            {
                client.ItemChanged += ClientOnItemChanged;
                int handle = client.SubscribeItem(serverDescriptor, itemDescriptor, requestedUpdateRate, testState);
                completed = completeEvent.WaitOne(waitTimeout);
                client.UnsubscribeItem(handle);
                client.ItemChanged -= ClientOnItemChanged;
            }

            TestSection.Assert();

            int count = argsList.Count;
            AssertUtilities.WriteLine($"count: {count}");
            AssertUtilities.WriteLine($"indexSet: {indexSet}");
            Assert.IsTrue(completed, "Not completed.");

            for (int i = 0; i < argsList.Count; i++)
            {
                object sender = senderList[i];
                Assert.IsNotNull(sender);
                Assert.AreSame(client, sender);

                EasyDAItemChangedEventArgs args = argsList[i];
                Assert.IsNotNull(args);

                Assert.AreEqual(testState, args.Arguments.State); // is correct for value types only
                Assert.IsNull(args.Arguments.Callback);

                Assert.AreEqual(new DAGroupParameters(requestedUpdateRate), args.Arguments.GroupParameters);
                Assert.AreEqual(itemDescriptor, args.Arguments.ItemDescriptor);
                Assert.AreEqual(serverDescriptor, args.Arguments.ServerDescriptor);

                if (!client.SupportsTechnology(serverDescriptor.Technology))
                {
                    Assert.IsFalse(args.Succeeded);
                    Assert.IsInstanceOfType(args.Exception, typeof(NotSupportedException));
                    continue;
                }

                Assert.IsNull(args.Exception);
                Assert.AreEqual(null, args.ErrorId);

                DAVtq actualVtq = args.Vtq;
                Assert.IsNotNull(actualVtq);
                AssertUtilities.WriteLine($@"{i}: {actualVtq}");

                object actualValue = actualVtq.Value;

                if ((indexSet == -1) || (i < indexSet))
                {
                    Assert.IsTrue(actualVtq.Quality.IsBad);
                    Assert.IsNull(actualValue);
                }
                else
                {
                    Assert.IsTrue(actualVtq.Quality.IsGood);
                    if (expectedType != null)
                    {
                        Assert.IsNotNull(actualValue);
                        Assert.AreEqual(expectedType, actualValue.GetType());
                    }
                }

                Assert.AreEqual(DateTimeKind.Utc, actualVtq.Timestamp.Kind);
                Assert.AreNotEqual(actualVtq.Timestamp, DateTime.MinValue);
                Assert.AreNotEqual(actualVtq.Timestamp, DateTime.MaxValue);
                Assert.IsTrue(actualVtq.Timestamp >= DateTime.FromFileTimeUtc(0));
                DateTime utcNow = DateTime.UtcNow;
                AssertUtilities.WriteLine($"UtcNow: {utcNow}");
                TimeSpan timeTolerance = TimeSpan.FromSeconds(60);
                Assert.IsTrue(actualVtq.Timestamp <= utcNow + timeTolerance, "actualVtq.Timestamp <= utcNow + timeTolerance");
            }
        }

        protected void SubscribeItemTest_Error(
            [NotNull] ServerDescriptor serverDescriptor,
            [NotNull] DAItemDescriptor itemDescriptor)
        {
            TestSection.Arrange();
            AssertUtilities.WriteLine(serverDescriptor);
            AssertUtilities.WriteLine(itemDescriptor);
            AssertUtilities.WriteLine("");

            TestEnvironment.PrecheckServer(serverDescriptor);
            EasyDAClient client = CreateClient();

            var senderList = new List<object>();
            var argsList = new List<EasyDAItemChangedEventArgs>();
            var completeEvent = new AutoResetEvent(initialState: false);

            void ClientOnItemChanged(object sender, EasyDAItemChangedEventArgs args)
            {
                AssertUtilities.WriteLine(args);
                senderList.Add(sender);
                argsList.Add(args);

                completeEvent.Set();
            }

            const int requestedUpdateRate = 100;
            const int testState = 777777777;

            TestSection.Act();
            client.ItemChanged += ClientOnItemChanged;
            int handle = client.SubscribeItem(serverDescriptor, itemDescriptor, requestedUpdateRate, testState);
            bool completed = completeEvent.WaitOne(5 * 1000);
            client.UnsubscribeItem(handle);
            client.ItemChanged -= ClientOnItemChanged;

            TestSection.Assert();

            AssertUtilities.WriteLine(argsList.Count);
            Assert.IsTrue(completed, "Not completed.");
            Assert.AreEqual(1, argsList.Count);

            {
                object sender = senderList[0];
                Assert.IsNotNull(sender);
                Assert.AreSame(client, sender);

                EasyDAItemChangedEventArgs args = argsList[0];
                Assert.IsNotNull(args);

                Assert.AreEqual(testState, args.Arguments.State); // is correct for value types only
                Assert.IsNull(args.Arguments.Callback);

                Assert.AreEqual(new DAGroupParameters(requestedUpdateRate), args.Arguments.GroupParameters);
                Assert.AreEqual(itemDescriptor, args.Arguments.ItemDescriptor);
                Assert.AreEqual(serverDescriptor, args.Arguments.ServerDescriptor);

                Exception exception = args.Exception;

                if (!client.SupportsTechnology(serverDescriptor.Technology))
                {
                    Assert.IsInstanceOfType(exception, typeof(NotSupportedException));
                    return;
                }

                Assert.IsNotNull(exception);
                AssertUtilities.WriteLine($"exception.Message: {exception.Message}");
                AssertUtilities.WriteLine($"args.ErrorId: {args.ErrorId}");
                Assert.IsNull(args.Vtq);

                AssertBaseException(exception,
                    -1073479673, // OPC_E_UNKNOWNITEMID
                    "http://opcfoundation.org/DataAccess/",
                    "E_UNKNOWN_ITEM_NAME");
            }
        }
    }
}
